/*
 * class Node
 *
 * Created on 20. jen 2006, 11:12
 *
 * @author Jiri Kremser
 * @version 1.0
 */

package ctverecku;
import java.lang.Math;


public class Node {
    public static final int DIM = 3;
    
    public static int[][] endDesk = {{1,2,3},
                                     {8,0,4},
                                     {7,6,5}};
    public static int[][] noSolutionDesk = {{-1,-1,-1},
                                            {-1,-1,-1},
                                            {-1,-1,-1}};
    public static Node endNode = new Node(Node.endDesk,null);
    public static Node noSolution = new Node(Node.noSolutionDesk,null);
    
    private int[][] desk = new int[DIM][DIM];
    private int d;
    private int h;
    private Node prevNode;
    
    private enum Direction {
        UP, DOWN, LEFT, RIGHT
    }
    
    public static Node[] visited = new Node[9*8*7*6*5*4*3*2];
    public static int visitedCount = 0;
//*****************************************************************************
            
    
    /**
     * Constructor
     */
    public Node(int[][] dsk,Node prvNode) {
        for(int i=0;i<DIM;i++){
            for(int j=0;j<DIM;j++){
                desk[i][j]=dsk[i][j];
            }
        }
        this.prevNode = prvNode;
        this.d = getD();
        this.h = getH();
    }
    /**
     * Constructor2
     */    
    public Node(Node node,Node prvNode) {
        this.desk = node.desk;
        this.prevNode = prvNode;
        this.d = getD();
        this.h = getH();
    }
    
    /**
     * isSame()
     */
    public boolean isSame(Node which) {
        for (int i = 0; i < DIM; i++) {
            for (int j = 0; j < DIM; j++) {
                if (this.desk[i][j] != which.desk[i][j]) {
                    return false;
                }
            }
        }
        return true;
    }
    
    
    /**
     * isEnd()
     */
    public boolean isEnd() {
        return this.isSame(endNode);
    }
    
    /**
     * isVisited()
     */
    public boolean isVisited() {
        for (int i = 0;i<Node.visitedCount;i++) {
            if (this.isSame(Node.visited[i])) {
                return true;
            }
        }
        return false;
    }
    
    /**
     * visit()
     */
    public void visit() {
        Node.visited[Node.visitedCount++] = this;
    }
    
    /**
     * clearVisited()
     */
    public static void clearVisited() {
        for (int i = 0; i < Node.visitedCount; i++) {
            Node.visited[Node.visitedCount] = null;
        }
        Node.visitedCount = 0;
    }
    
    
    /**
     * move()
     */
    public Node move(int Hx, int Hy, Direction where) {
        Node nod = new Node(this.desk,this.prevNode);
        switch (where) {
            case UP:
                nod.desk[Hy][Hx] = nod.desk[Hy-1][Hx];
                nod.desk[Hy-1][Hx] = 0;
                break;
            case DOWN:
                nod.desk[Hy][Hx] = nod.desk[Hy+1][Hx];
                nod.desk[Hy+1][Hx] = 0;
                break;
            case LEFT:
                nod.desk[Hy][Hx] = nod.desk[Hy][Hx-1];
                nod.desk[Hy][Hx-1] = 0;
                break;
            case RIGHT:
                nod.desk[Hy][Hx] = nod.desk[Hy][Hx+1];
                nod.desk[Hy][Hx+1] = 0;
                break;
        }
        return nod;
    }
    
    
    /**
     * getPrevNode()
     */
    public Node getPrevNode() {
        return prevNode;
    }
    
    /**
     * setPrevNode()
     */
    public void setPrevNode(Node prvNode) {
        this.prevNode = prvNode;
    }        
    
    
    /**
     * selectBest()
     *
     *  way - type of heuristic (true - Manhattan)
     */
    public static Node selectBest(Node[] kids, boolean way, boolean checkVisited) {
        Node[] uberKids = new Node[4];
        if (kids == null) {
            return Node.noSolution;
        }
        
        int ind = 0,akt = 0,min = 33, max = 0,order = 0;
        
        if (checkVisited) {
            for (int i=0;i<kids.length;i++) {
                if (kids[i] != null) {
                    akt = (way) ? kids[i].getD() : kids[i].getH();
                    if (akt < min) { 
                        min = akt;        
                        ind = i;
                    }
                }                     
            }
            for (int j=0;j<kids.length;j++) {
                if (kids[j] != null) {
                    akt = (way) ? kids[j].getD() : kids[j].getH();
                }
                if (akt == min) {
                    uberKids[order++] = kids[j];
                }                
            }
            ind = ( (int) ( Math.random()*1000)%order);

            if (uberKids[ind] != null) return uberKids[ind];
            else {
                return Node.noSolution;
            }
        }
        else {
            ind = ( (int) ( Math.random()*1000)%kids.length);
            while (kids[ind] == null) {
                ind = ( (int) ( Math.random()*1000)%kids.length);
            }   
            if (kids[ind] != null) {
                return kids[ind];
            }
            else {
                return Node.noSolution;
            }
        }                        
    }
        
    
    /**
     * children()
     */
    public Node[] children(boolean checkVisited) {
        Node[] kids = new Node[4];
        int holeX = 0, holeY = 0,k = 0;
        for(int i=0;i<DIM;i++){  // find the hole
            for(int j=0;j<DIM;j++){
                if (desk[i][j] == 0) {
                    holeX = j;
                    holeY = i;
                }
            }
        }
        switch (holeX) {
            case 0:
                if ((checkVisited) && (!this.move(holeX,holeY,Direction.RIGHT).isVisited())) {
                    kids[k++] = this.move(holeX,holeY,Direction.RIGHT);
                }
                else {
                    if (!checkVisited) {
                        kids[k++] = this.move(holeX,holeY,Direction.RIGHT);
                    }                    
                }
                break;
            case 1:
                if ((checkVisited) && (!this.move(holeX,holeY,Direction.RIGHT).isVisited())) {
                    kids[k++] = this.move(holeX,holeY,Direction.RIGHT);
                }
                else {
                    if (!checkVisited) {
                        kids[k++] = this.move(holeX,holeY,Direction.RIGHT);
                    }                
                }
                if ((checkVisited) && (!this.move(holeX,holeY,Direction.LEFT).isVisited())) {
                    kids[k++] = this.move(holeX,holeY,Direction.LEFT);
                }
                else {
                    if (!checkVisited) {
                        kids[k++] = this.move(holeX,holeY,Direction.LEFT);
                    }                
                }
                break;
            case 2:
                if ((checkVisited) && (!this.move(holeX,holeY,Direction.LEFT).isVisited())) {
                    kids[k++] = this.move(holeX,holeY,Direction.LEFT);
                }
                else {
                    if (!checkVisited) {
                        kids[k++] = this.move(holeX,holeY,Direction.LEFT);
                    }                
                }
                break;
        }
        switch (holeY) {
            case 0:
                if ((checkVisited) && (!this.move(holeX,holeY,Direction.DOWN).isVisited())) {
                    kids[k++] = this.move(holeX,holeY,Direction.DOWN);
                }
                else {
                    if (!checkVisited) {
                        kids[k++] = this.move(holeX,holeY,Direction.DOWN);
                    }                
                }
                break;
            case 1:
                if ((checkVisited) && (!this.move(holeX,holeY,Direction.DOWN).isVisited())) {
                    kids[k++] = this.move(holeX,holeY,Direction.DOWN);
                }
                else {
                    if (!checkVisited) {
                        kids[k++] = this.move(holeX,holeY,Direction.DOWN);
                    }                
                }
                if ((checkVisited) && (!this.move(holeX,holeY,Direction.UP).isVisited())) {
                    kids[k++] = this.move(holeX,holeY,Direction.UP);
                }
                else {
                    if (!checkVisited) {
                        kids[k++] = this.move(holeX,holeY,Direction.UP);
                    }
                }
                break;
            case 2:
                if ((checkVisited) && (!this.move(holeX,holeY,Direction.UP).isVisited())) {
                    kids[k++] = this.move(holeX,holeY,Direction.UP);
                }
                else {
                    if (!checkVisited) {
                        kids[k++] = this.move(holeX,holeY,Direction.UP);
                    }
                }
                break;
        }
        Node[] kidsX=null;
        if (k == 0) {
            // no solution
        } else {
            kidsX = new Node[k];
            
            int m = 0;
            for (int l = 0;l<k;l++) {
                if (!kids[l].isVisited()) {
                    kidsX[m++] = kids[l];
                }
            }
            System.out.println("non-repeating moves: " + k + " \n");
            System.out.println("                 |");
            System.out.println("                \\|/");
            System.out.println("                 V\n");
                        
        }        
        return kidsX;
    }
    
    
    
    /**
     * getD()
     *
     * Manhattan distance - heuristic
     */
    public int getD() {
        d=0;
        for(int i=0;i<DIM;i++){
            for(int j=0;j<DIM;j++){
                switch(desk[i][j]){
                    case 1:
                        d += i + j;
                        break;
                    case 2:
                        d += i + Math.abs(j - 1);
                        break;
                    case 3:
                        d += i + Math.abs(j - 2);
                        break;
                    case 4:
                        d += Math.abs(i - 1)+Math.abs(j-2);
                        break;
                    case 5:
                        d += Math.abs(i - 2)+Math.abs(j-2);
                        break;
                    case 6:
                        d += Math.abs(i - 2)+Math.abs(j-1);
                        break;
                    case 7:
                        d += Math.abs(i - 2) + j;
                        break;
                    case 8:
                        d += Math.abs(i - 1) + j;
                        break;
                }
            }
        }
        return d;
    }
    
    
    /**
     * getH()
     *
     * Number of wrong - heuristic
     */
    public int getH() {
        h=0;
        for(int i=0;i<DIM;i++){
            for(int j=0;j<DIM;j++){
                if ((this.desk[i][j] != 0) && (this.desk[i][j] != endDesk[i][j])) {
                    h++;
                }
            }
        }
        return h;
    }


    /**
     * toString()
     */    
    public String toString() {
        String output = "";
        for(int i=0;i<DIM;i++){
            output += "        " + ( (desk[i][0] == 0)?" ":desk[i][0] ) + " | " 
                   + ( (desk[i][1] == 0)?" ":desk[i][1] )+ " | " 
                   + ( (desk[i][2] == 0)?" ":desk[i][2] ) + "\n";
            if (i != DIM - 1 ) output += ("       -----------\n");
        }
        output += "First heuristic:" + this.getH() + "\nManhattan heuristic: " + this.getD();
        return output;
    }
}